"""
A library to extract and inspect a Symbian ROM Image. It is based on the information
from the Symbian 7.0(s) header files and information in SDL_7.0.chm.
"""

__version__ = "$Revision: 1.11 $"
__copyright__ = "Copyright 2005, Job de Haas <job@itsx.com>"

__all__ = ["Rom", "ExtensionRom", "ExportBase", "ExportEntry", "RomEntry", "RomImageHeader", "CommonHeader", "E32ImageHeader", "DllRefTable", "DllRefTableEntry", "RomDir" ,"RootDirInfo", "RootDirList", "TRomHeaderSize", "TRomImageHeaderSize"]

import sys
import os
from mmap import *
import struct, sstruct
import string
import types
import codecs

SEEK_BEG = 0
SEEK_SET = 1
SEEK_END = 2

TRomHeader = """
    <  # little endian
    iMajor:                 B       # ROM image major version.
    iMinor:                 B       # ROM image minor version.
    iBuild:                 H       # ROM image build number.
    iTime:                  8s      # ROM image build datetime.
    iRomBase:               L       # ROM virtual base address.
    iRomSize:               L       # ROM image size in bytes (including the second section for
                                    # sectioned ROMs).
    iRomRootDirectoryList:  L       # Virtual address of the TRomRootDirList structure.
    iKernDataAddress:       L       # Kernel data virtual address (when in RAM).
    iKernStackAddress:      L       # Kernel stack virtual address (when in RAM).
    iPrimaryFile:           L       # Virtual address of (first) primary's TRomImageHeader.
    iSecondaryFile:         L       # Virtual address of secondary's TRomImageHeader.
    iCheckSum:              L       # 32-bit checksum of the image
    iLanguage:              8s      # a bitmask identifying which of the languages are supported by
                                    # the ROM image
    iHardware:              L       # unique number identifying the hardware for which this ROM image
                                    # has been built.
    iScreenWidth:           L       # Screen width in pixels
    iScreenHeight:          L       # Screen height in pixels
    iScreenBitsPerPixel:    L       # Maximum bits per pixel
    iRomSectionHeader:      L       # a pointer to a TRomSectionHeader, if the ROM is sectioned
    iTotalSvDataSize:       L       # the amount of memory mapped for the Kernel data chunk
    iVariantFile:           L       # Address of variant file TRomEntry
    iExtensionFile:         L       # Address of TRomEntry for the first extension file
    iRelocInfo:             L       # Address of the first area to be relocated into RAM
    iTraceMask:             L       # Initial value for the Kernel trace mask.
    iUserDataAddress:       L       # used by the ESplitSingle process model
    iTotalUserDataSize:     L       # used by the ESplitSingle process model
    iDebugPort:             L       # Value interpreted by the Bootstrap and Kernel as the debug port to use
    iPad:                   24s
"""

TRomHeaderSize = sstruct.calcsize(TRomHeader)
assert TRomHeaderSize == 128

TExtensionRomHeader = """
    <  # little endian
    iMajor:                 B       # ROM image major version.
    iMinor:                 B       # ROM image minor version.
    iBuild:                 H       # ROM image build number.
    iTime:                  8s      # ROM image build datetime.
    iRomBase:               L       # ROM virtual base address.
    iRomSize:               L       # ROM image size in bytes (including the second section for
    iRomRootDirectoryList:  L       # Virtual address of the TRomRootDirList structure.
    iCheckSum:              L       # 32-bit checksum of the image
    iKernelMajor:           B       # Kernel image major version.
    iKernelMinor:           B       # Kernel image minor version.
    iKernelBuild:           H       # Kernel image build number.
    iKernelTime:            8s      # Kernel image build datetime.
    iKernelCheckSum:        L       # Kernel 32-bit checksum of the image
    iPad:                   84s
"""

TExtensionRomHeaderSize = sstruct.calcsize(TExtensionRomHeader)
assert TExtensionRomHeaderSize == 128


TRomRootDirectoryList = """
    <  # little endian
    iNumRootDirs:   L        # Number of RootDirInfo entries
"""

TRomRootDirectoryListSize = sstruct.calcsize(TRomRootDirectoryList)
assert TRomRootDirectoryListSize == 4

TRootDirInfo = """
    <  # little endian
    iHardwareVariant:   L
    iAddressLin:        L
"""

TRootDirInfoSize = sstruct.calcsize(TRootDirInfo)
assert TRootDirInfoSize == 8

TRomEntry = """
    <  # little endian
    iSize:          L        # Size of file.
    iAddressLin:    L        # Virtual address of entry
    iAtt:           B 
    iNameLength:    B 
"""

TRomEntrySize = sstruct.calcsize(TRomEntry)
assert TRomEntrySize == 10

TRomDir = """
    <  # little endian
    iSize:     L            # 
"""

TRomDirSize = sstruct.calcsize(TRomDir)
assert TRomDirSize == 4

TCommonHeader = """
    <  # little endian
    iUid1:              L 
    iUid2:              L 
    iUid3:              L 
    iCheck:             L 
    iSignature:         L       # 'EPOC'
"""

TCommonHeaderSize = sstruct.calcsize(TCommonHeader)
assert TCommonHeaderSize == 20

TRomImageHeader = """
    <  # little endian
    iUid1:              L 
    iUid2:              L 
    iUid3:              L 
    iCheck:             L 
    iEntryPoint:        L     # Entrypoint of this executable (offset within this file).
    iCodeAddress:       L 
    iDataAddress:       L 
    iCodeSize:          L 
    iTextSize:          L 
    iDataSize:          L 
    iBssSize:           L 
    iHeapSizeMin:       L 
    iHeapSizeMax:       L 
    iStackSize:         L 
    iDllRefTable:       L     # Address of DLL reference table (a list of the DLLs which this executable needs to run).
                              # This is a TDllRefTable structure.
    iExportDirCount:    L     # Number of functions exported by this executable.
    iExportDir:         L     # Export directory address. The export directory is a simple list of functions, 
                              # which can be indexed by ordinal
    iCheckSumCode:      L 
    iCheckSumData:      L 
    iMajor:             B     # file image major version.
    iMinor:             B     # file image minor version.
    iBuild:             H     # file image build number.
    iFlags:             L 
    iPriority:          L     # Priority of this process
    iDataBssLinearBase: L     # Data and BSS linear base address, where this process expects its data chunk to be when it runs.
    iNextExtension:     L     # Address of the TRomEntry for the next extension file.
                              # This field is only used if there is more than one extension.
                              # The first extension is found using the TRomHeader
    iHardwareVariant:   L     # A number denoting the hardware Variant,
                              # used to determine if this executable can run on any particular system
"""

TRomImageHeaderSize = sstruct.calcsize(TRomImageHeader)
assert TRomImageHeaderSize == 100

TE32ImageHeader = """
    <  # little endian
    iUid1:              L 
    iUid2:              L 
    iUid3:              L 
    iCheck:             L 
    iSignature:         L       # EPOC
    iCpu:               L       # 0x1000 = X86, 0x2000 = ARM, 0x4000 = M*Core
    iCheckSumCode:      L       # sum of all 32 bit words in .text
    iCheckSumData:      L       # sum of all 32 bit words in .data
    iMajor:             B       # major version.
    iMinor:             B       # minor version.
    iBuild:             H       # build number.
    iTime:              8s
    iFlags:             L       # 0 = exe, 1 = dll, +2 = no call entry points
    iCodeSize:          L       # size of code, import address table, constant data and export dir
    iDataSize:          L       # size of initialized data
    iHeapSizeMin:       L
    iHeapSizeMax:       L
    iStackSize:         L
    iBssSize:           L
    iEntryPoint:        L       # offset into code of entry point
    iCodeBase:          L       # where the code is linked for
    iDataBase:          L       # where the data is linked for
    iDllRefTableCount:  L       # filling this in enables E32ROM to leave space for it
    iExportDirOffset:   L       # offset into the file of the export address table
    iExportDirCount:    L
    iTextSize:          L       # size of just the text section
    iCodeOffset:        L       # file offset to code section
    iDataOffset:        L       # file offset to data section
    iImportOffset:      L       # file offset to import section
    iCodeRelocOffset:   L       # relocations for code and const
    iDataRelocOffset:   L       # relocations for data
    iPriority:          L       # priority of this process
"""

TE32ImageHeaderSize = sstruct.calcsize(TE32ImageHeader)
assert TE32ImageHeaderSize == 124

TDllRefTable = """
    <  # little endian
    iFlags:             H
    iNumberOfEntries:   H
"""

TDllRefTableSize = sstruct.calcsize(TDllRefTable)
assert TDllRefTableSize == 4

TDllRefTableEntry = """
    <  # little endian
    iEntryPoint:        L
    iDllRefTable:       L
"""

TDllRefTableEntrySize = sstruct.calcsize(TDllRefTableEntry)
assert TDllRefTableEntrySize == 8

def mapfile(filename):
    file = open(filename, "r+")
    size = os.path.getsize(filename)
    return mmap(file.fileno(), size)

# as I have no clue about python yet, I just create a container object to hold the indices
class ExportBase:
    def __init__(self):
        self.byaddr = {}
        self.byfunc = {}

class ExportEntry:
    """Represents the information on a exported entry.
    """
    def __init__ (self, address, funcaddr, ordinal, filename, fullname, funcname, mangled=None):
        self.address = address
        self.funcaddr = funcaddr
        self.filename = filename
        self.fullname = fullname
        self.ordinal = ordinal
        self.funcname = funcname
        self.mangled = mangled

class Rom:
    """The basic Rom object representing a ROM image.

    This class unpacks the main ROM Header. This can then be used to further unpack
    individual files and directories.
    """
    def __init__ (self, filename):
        """
        The filename points to a Symbian ROM image file.
        """
        self.filename = filename
        # Check if file exists and if so ask if overwrite necessary.
        if not os.access(self.filename,os.O_RDWR):
            print "Aborting, file %s does not exist..." % self.filename
            return None
        # Open and memory map the file.
        self.file = mapfile(filename)
        sstruct.unpack(TRomHeader, self.file[128:128+TRomHeaderSize], self)
        header = sstruct.pack(TRomHeader, self)

class ExtensionRom:
    """The extension Rom object representing am extended ROM image.

    This class unpacks the main ROM Header. This can then be used to further unpack
    individual files and directories.
    """
    def __init__ (self, filename):
        """
        The filename points to a Symbian Extension ROM image file.
        """
        self.filename = filename
        # Check if file exists and if so ask if overwrite necessary.
        if not os.access(self.filename,os.O_RDWR):
            print "Aborting, file %s does not exist..." % self.filename
            return None
        # Open and memory map the file.
        self.file = mapfile(filename)
        sstruct.unpack(TExtensionRomHeader, self.file[0:TExtensionRomHeaderSize], self)
        header = sstruct.pack(TExtensionRomHeader, self)

class RomEntry:
    """This is the basic item of the Symbian ROM filesystem.

    Normally it consists of TRomImageHeader files, but E32ImageHeaders have also been encountered in the ROM.
    """
    def __init__ (self, rom, address, parent=None):
        """ Pass a Rom object and the start address in the local ROM space.
        """
        self.rom = rom
        self.file = rom.file
        self.parent = parent
        self.address = address
        end = self.address+TRomEntrySize
        sstruct.unpack(TRomEntry, self.file[self.address:end], self)
        self.file.seek(end)
        utfreader = codecs.getreader('utf-16-le')
        namereader = utfreader(self.file)
        self.name = namereader.read(self.iNameLength*2)
        self.end = (end + self.iNameLength*2 + 3) & ~3  # compensate for alignment
        self.size = self.end - self.address
        if self.iAtt & 0x10:
            self.romdir = RomDir(rom,self.iAddressLin - self.rom.iRomBase, self)
        else:
            address = self.iAddressLin - rom.iRomBase
            if 0 < address < self.rom.iRomSize:
                common = CommonHeader(self.file, self.iAddressLin - rom.iRomBase)
                # a bit of a hack to deal with a case where a normal E32Image was found in a ROM
                if (common.iSignature == 0x434f5045):
                    self.fileheader = E32ImageHeader(self.file, self.iAddressLin - rom.iRomBase)
                    self.fileheader.iDllRefTable = 0x0  # add to be able to prevent loading
                    # Create ExportDir similar to RomImageHeader
                    self.fileheader.iExportDir = self.fileheader.iExportDirOffset + self.iAddressLin
                else:
                    self.fileheader = RomImageHeader(self.file, self.iAddressLin - rom.iRomBase)
                    self.fileheader.iSignature = 0x0    # add to be able to distinguish
                if ( self.fileheader.iUid1 == 0x10000079) | ( self.fileheader.iUid1 == 0x1000007A):
                    if (self.fileheader.iDllRefTable != 0 ):
                        if (0 < self.fileheader.iDllRefTable-rom.iRomBase < rom.iRomSize):
                            self.dll = DllRefTable(self.file, self.fileheader.iDllRefTable-rom.iRomBase)
                        else:
                            print 'dllref address %d out of range (size 0x%08x, %s)'%(self.fileheader.iDllRefTable-rom.iRomBase, self.rom.iRomSize, self.name)
                    if (self.fileheader.iExportDirCount > 0):
                        self.fileheader.iExports = []
                        for export in range(self.fileheader.iExportDirCount):
                            #address = self.fileheader.iExportDir+export*4 - rom.iRomBase - 0x1b0
                            address = self.fileheader.iExportDir+export*4 - rom.iRomBase
                            exportaddr, = struct.unpack('L',self.file[address:address+4])
                            self.fileheader.iExports.append(exportaddr)
                        #print "exportdircount %d"%self.fileheader.iExportDirCount
                        #if (self.name == 'EUser.dll'):
                        #    print "Rombase %x"%rom.iRomBase
                        #    print "linaddr %x"%self.iAddressLin
                        #    for export in range(self.fileheader.iExportDirCount):
                        #        address = self.fileheader.iExportDir+export*4 - rom.iRomBase
                        #        print '%d 0x%08x 0x%08x'%(export,self.fileheader.iExports[export],address)
            else:
                print 'address 0x%08x out of range (size 0x%08x)'%(address, self.rom.iRomSize)

    def path(self):
        path = ""
        parent = self.parent
        while parent != None:
            path = parent.name + '/' + path
            parent = parent.parent
        path = '/' + path
        return path

    def fullname(self):
        return self.path() + self.name

class CommonHeader:
    """This extracts the common part of the header for a Symbian exe, dll or similar. Both in ROM or normal.
    """
    def __init__ (self, file, address=0):
        """ Pass an opened file object and the start address in that file.
        Normally this is the start of the file (0).
        But this can also be used with a Rom object and the appropriate offset
        """
        self.file = file
        self.address = address
        self.file.seek(address)
        self.end = self.address+TCommonHeaderSize
        data = self.file.read(TCommonHeaderSize)
        sstruct.unpack(TCommonHeader, data, self)

class RomImageHeader:
    """This extracts the header for a Symbian ROM exe, dll or similar.
    """
    def __init__ (self, file, address=0):
        """ Pass an opened file object and the start address in that file.
        Normally this is the start of the file (0).
        But this can also be used with a Rom object and the appropriate offset
        """
        self.file = file
        self.address = address
        self.file.seek(address)
        self.end = self.address+TRomImageHeaderSize
        data = self.file.read(TRomImageHeaderSize)
        sstruct.unpack(TRomImageHeader, data, self)

class E32ImageHeader:
    """This extracts the header for a normal Symbian exe, dll or similar.
    """
    def __init__ (self, file, address=0):
        """ Pass an opened file object and the start address in that file.
        Normally this is the start of the file (0).
        But this can also be used with a Rom object and the appropriate offset
        """
        self.file = file
        self.address = address
        self.file.seek(address)
        self.end = self.address+TE32ImageHeaderSize
        data = self.file.read(TE32ImageHeaderSize)
        sstruct.unpack(TE32ImageHeader, data, self)

class DllRefTableEntry:
    def __init__ (self, file, address):
        self.file = file
        self.address = address
        self.file.seek(address)
        data = self.file.read(TDllRefTableEntrySize)
        sstruct.unpack(TDllRefTableEntry, data, self)

class DllRefTable:
    def __init__ (self, file, address):
        self.file = file
        self.address = address
        self.file.seek(address)
        data = self.file.read(TDllRefTableSize)
        sstruct.unpack(TDllRefTable, data, self)
        self.DllEntries = []
        start = address + TDllRefTableSize
        for i in range(self.iNumberOfEntries):
            self.DllEntries.append(DllRefTableEntry(file, start + i * TDllRefTableEntrySize))

class RomDir:
    def __init__ (self, rom, address, romentry = None):
        self.rom = rom
        self.file = rom.file
        self.address = address
        if (romentry is None):
            self.name = "Root"
        else:
            self.name = romentry.name
        sstruct.unpack(TRomDir, self.file[self.address:self.address+TRomDirSize], self)
        start = self.address+TRomDirSize;
        ptr = start
        end = start+self.iSize
        self.dir = []
        while ptr < end:
            self.dir.append(RomEntry(rom, ptr, romentry))
            ptr += self.dir[-1].size

class RootDirInfo:
    def __init__ (self, rom, address):
        self.rom = rom
        self.file = rom.file
        self.address = address
        sstruct.unpack(TRootDirInfo, self.file[self.address:self.address+TRootDirInfoSize], self)
        #print 'iHardwareVariant: 0x%0.8x' % self.iHardwareVariant
        #print 'iAddressLin: 0x%0.8x' % self.iAddressLin
        start = self.iAddressLin - self.rom.iRomBase
        self.rootdir = RomDir(self.rom,start)

class RootDirList:
    def __init__ (self, rom, address):
        self.rom = rom
        self.file = rom.file
        self.address = address
        sstruct.unpack(TRomRootDirectoryList, self.file[self.address-self.rom.iRomBase:self.address-self.rom.iRomBase+TRomRootDirectoryListSize], self)
        #print 'iNumRootDirs: 0x%0.8x' % self.iNumRootDirs
        start = self.address - self.rom.iRomBase + TRomRootDirectoryListSize
        #print 'start rootdirinfo %x' % start
        self.rootdirinfo = RootDirInfo(self.rom,start)


# Main
if __name__ == '__main__':
    print "This is a library"

